% @author Victoria Steck

:-dynamic available/1.
:-dynamic beingfixed/1.
:-dynamic secondunfixed/1.
:-dynamic tied/2.
:-dynamic rules/2.
:-dynamic numinferences/1.
:-dynamic beingconsidered/1.
:-dynamic inferencecolors/1.
:-dynamic numtied/1.
:-dynamic guessnum/1.

% the main method to give information to the user about which colors can be used
% This method calls main to generate a guess
go :- 	write('Welcome to Mastermind.'), nl,
	write('Please think of a five digit code using the numbers 1-8. Type yes. when you are ready: '),
	nl, read(yes),
	main.

% This method first checks to see if all colors have been tied and writes the code if so.
% If not it gets a new value for being considered, generates a new guess and calls the execute method.	
main :- numtied(X), X is 5, gen_guess(5, [], Result),
	write('Your secret code is'), write(Result), clean.
main :- inferencecolors(X),
	getbeingconsidered(X),
	guesses(Result, Y, Z),
	execute(Result, Y, Z).

% The execute method first checks to see if the number of bulls is 5.
% If not it calls methods to create rules based on the number of bulls, make inferences based on number of cows and get a new value for being fixed if necessary.
execute(_, 5, _) :- write('Thanks for playing!'), clean.
execute(Guess, Bulls, Cows) :- handlebulls(Bulls, Cows, Guess),
			       numcows(Cows),
			       getbeingfixed(Bulls),
			       main.

% This method calls the gen_guess method to generate a new guess with guess number and gets the number of bulls and cows from the user.
guesses(Result, Y, Z) :- retract(guessnum(X)), 
		   write('Guess Number '), write(X), nl,
		   A is X + 1, assert(guessnum(A)),
		   gen_guess(5, [], Result), write('My guess is: '),
		   write(Result), nl,
		   write('Please enter the number of bulls followed by a period: '),
        	   read(Y), nl,
        	   write('Please enter the number of cows followed by a period: '),
        	   read(Z), nl.
	
%This method uses what we already know about rules and number of colors tied to generate new guesses
gen_guess(0, Guess, Guess).
gen_guess(Pos, Temp, Guess) :- tied(X, Pos),
			 Pos2 is Pos - 1,
			 gen_guess(Pos2, [X|Temp], Guess).
gen_guess(Pos, Temp, Guess) :- not(beingfixed(0)),
			 beingfixed(X),
			 rules(X, [H|_]),
			 H is Pos,
			 Pos2 is Pos-1,
			 gen_guess(Pos2, [X|Temp], Guess).
gen_guess(Pos, Temp, Guess) :- beingconsidered(X), 
			 Pos2 is Pos - 1,
			 gen_guess(Pos2, [X|Temp], Guess).

%This method is used by cleanup to remove all instances of a given position from a given rule
rempos(_,[],[]).
rempos(Pos, [Pos|T], Result) :- rempos(Pos, T, Result).
rempos(Pos, [H|T], [H|TResult]) :- rempos(Pos, T, TResult).

% This method removes the first instance of a given color from the inference list.
remfirst(_, [], []).
remfirst(Pos, [Pos|T], T).
remfirst(Pos, [_|T], Result) :- remfirst(Pos, T, Result).

% This method removes a given position from all rules of a given color.
remall(Color, Pos) :- retract(rules(Color, X)),
		      rempos(Pos, X, Result),
		      remall(Color, Pos),
		      asserta(rules(Color, Result)).
remall(_, _).

%This method removes a given position from all rules in the database
cleanup(Pos) :- retract(rules(X,Y)),
		rempos(Pos, Y, Result),
		cleanup(Pos),
		asserta(rules(X,Result)).
cleanup(_).

% ties colors to positions and removes those positions from all other rules
cleanupall :- rules(X,[H|T]),
	      length([H|T], 0, Length),
	      Length is 1, 
	      retract(rules(X, [H|T])),
	      assert(tied(X,H)),
              retract(numtied(Z)),
              R is Z + 1,
              assert(numtied(R)),
	      retract(inferencecolors(Y)),
	      rempos(X, Y, Result),
	      assert(inferencecolors(Result)),
	      cleanup(H),
	      cleanupall.
cleanupall.

% This method gets the next color for being considered.  If five inferences have not been made it takes colors
% from the available list.  Otherwise it takes the next color from the inference list.
getbeingconsidered([H|T]) :- numinferences(X), X is 5,
		      beingfixed(Y),
		      H=Y,
		      getbeingconsidered(T).
getbeingconsidered([H|_]) :- numinferences(X), X is 5,
		      retract(beingconsidered(_)),
		      assert(beingconsidered(H)).		      
getbeingconsidered(_) :- retract(available([H|T])),
		      retract(beingconsidered(_)),
		      assert(available(T)),
		      assert(beingconsidered(H)).		

% This method assigns a new value to beingfixed if it is 0 and the inference list is not empty.
getbeingfixed(_) :- not(beingfixed(0)).
getbeingfixed(0).
getbeingfixed(_) :- inferencecolors([]),
		    retract(beingfixed(_)),
		    assert(beingfixed(0)).
getbeingfixed(_) :- retract(inferencecolors([H|T])),
		 	retract(beingfixed(0)),
		 	assert(beingfixed(H)),
			assert(inferencecolors(T)).

% This checks the length of a given list and is used by the cleanup method to tie colors to positions if their position list only has one value.
length([], Length, Length).
length([_|T], Temp, Length) :- Temp2 is Temp + 1, length(T, Temp2, Length).

% This method is used to make inferences.  If the number of cows is 0 and there is a value being fixed than the value is tied to the given color.
% If there is one cow then the first position is deleted from the list of rules.  If there are two cows then fix the position of being considered
% to the current position of beingfixed.
numcows(0) :- beingfixed(0).
numcows(0) :- inferencecolors([]),
	      retract(beingfixed(X)),
	      retract(rules(X, [H|_])),
	      assert(tied(X, H)),
	      retract(numtied(Z)),
	      Y is Z + 1,
	      assert(numtied(Y)),
	      cleanup(H),
	      cleanupall,
	      assert(beingfixed(0)).
numcows(0) :- retract(beingfixed(X)),
              retract(rules(X, [H|_])),
              assert(tied(X, H)),
	      retract(numtied(Z)),
              Y is Z + 1,
              assert(numtied(Y)), 
              cleanup(H),
              cleanupall,
              assert(beingfixed(0)). 
numcows(1) :- beingfixed(X),
	      rules(X,[H|_]),
	      remall(X,H).
numcows(2) :- numinferences(X),
	      not(X is 5),
	      Y is X + 1,
	      retract(numinferences(X)),
	      assert(numinferences(Y)),
	      beingfixed(Z),
	      rules(Z, [H|_]),
	      beingconsidered(A),
              assert(tied(A, H)),
	      retract(numtied(B)),
	      C is B+1,
	      assert(numtied(C)),
              cleanup(H),
	      cleanupall.
numcows(2) :- beingfixed(X),
	      rules(X, [H|_]),
	      beingconsidered(Y),
	      assert(tied(Y, H)),
	      retract(rules(Y,_)),
	      retract(inferencecolors(Z)),
	      remfirst(Y, Z, Result),
	      assert(inferencecolors(Result)),
	      cleanup(H),
	      cleanupall.

% This method is responsible for adding rules based on gained inferences
handlebulls(_, _, _) :- numinferences(X),
			X is 5.
handlebulls(Bulls, _, _) :- numtied(X),
				Numrules is Bulls - X - 1,
				Numrules is -1.
handlebulls(Bulls, 0, Guess) :- not(beingfixed(0)),
			 	retract(numinferences(X)),
				numtied(Z),
			 	Numrules is Bulls - Z - 1,
			 	Inferences is X + Numrules,
			 	assert(numinferences(Inferences)),
			 	handleaddrules(Numrules, Guess).
handlebulls(Bulls, 2, Guess) :-retract(numinferences(X)),
				numtied(Z),
                                Numrules is Bulls - Z,
                                Inferences is X + Numrules,
                                assert(numinferences(Inferences)),
                                handleaddrules(Numrules, Guess).
handlebulls(Bulls, _, Guess) :- retract(numinferences(X)),
				numtied(Z),
				Numrules is Bulls - Z,
				Inferences is X + Numrules,
				assert(numinferences(Inferences)),
				handleaddrules(Numrules, Guess).			

% This method adds positions to a new rule based on positions in the guess.
handleaddrules(0, _).
handleaddrules(Numrules, Guess) :- beingconsidered(Z),
				   addlists(Z, 1, Guess, []),
				   Numrules2 is Numrules - 1,
				   handleaddrules(Numrules2, Guess).			    

addlists(Color, 6, _, Result) :- reverse(Result, Result2),
				 assert(rules(Color, Result2)),
				 retract(inferencecolors(X)),
				 append(X, [Color], Y),
				 assert(inferencecolors(Y)).
addlists(Color, Pos, [H|T], Result) :- Color=H,
				       Pos2 is Pos + 1,
				       addlists(Color, Pos2, T, [Pos|Result]).
addlists(Color, Pos, [_|T], Result) :- Pos2 is Pos + 1,
				       addlists(Color, Pos2, T, Result).				       	

% This method reverses a list and is used by addlists so that the positions are added to the rule in their proper order.
reverse(List, Reversed) :-
          reverse(List, [], Reversed).
reverse([], Reversed, Reversed).
reverse([Head|Tail], SoFar, Reversed) :-
          reverse(Tail, [Head|SoFar], Reversed).

% Gets the position of a given element in the list.
posnum(X, [H|_], Y, Y) :- X=H.
posnum(X, [_|T], Y, Z) :- R is Y+1, posnum(X, T, R, Z). 

%dynamic variables

% The list of available colors to make inferences from
available([1, 2, 3, 4, 5, 6, 7, 8]).
% This list of inference colors
inferencecolors([]).

% The number of inferences made
numinferences(0).
% The number of colors tied to position
numtied(0).
% The current color being fixed
beingfixed(0).
% The current color being considered
beingconsidered(0).
% The current guess number
guessnum(1).

% This method cleans up all the variables so the user doesn't have to restart prolog every time they want to play
clean :- retractrules,
	 retractied,
	 retract(guessnum(_)),
	 assert(guessnum(0)),
	 retract(available(_)),
	 assert(available([1, 2, 3, 4, 5, 6, 7, 8])),
	 retract(inferencecolors(_)),
	 assert(inferencecolors([])),
	 retract(numinferences(_)),
	 assert(numinferences(0)),
	 retract(numtied(_)),
	 assert(numtied(0)),
	 retract(beingfixed(_)),
	 assert(beingfixed(0)),
	 retract(beingconsidered(_)),
	 assert(beingconsidered(0)).

% This method retracts all rules as part of cleaning up.
retractrules :- retract(rules(_,_)),
		retractrules.
retractrules.

retractied :- retract(tied(_,_)),
	      retractied.
retractied.
