% @author Victoria Steck

:-dynamic filledin/1.
:-dynamic numtied/1.
:-dynamic available/1.
:-dynamic numinferences/1.
:-dynamic colors/1.
:-dynamic others/1.
:-dynamic beingconsidered/1.
:-dynamic beingfixed/1.
:-dynamic guessnum/1.
:-dynamic tied/2.
:-dynamic rules/2.
:-dynamic curpos/1.
:-dynamic numcorrect/1.
:-dynamic numadd/1.

% the main method to give information to the user about which colors can be used
% This method calls main to generate a guess
go :-   write('Welcome to Mastermind.'), nl,
        write('Please think of a five digit code using the numbers 1-8. Type yes. when you are ready: '),
        nl, read(yes), clean,
        main.

% This is the initial guessing method it generates guesses two colors at a time until it goes through all colors
% or gets five inferences.  When one of these two conditions is met it moves to phase 2.
main :- guessnum(5),
	phase2.
main :- numinferences(5),
	phase2.
main :- guesses(Result, Y, Z),
        execute(Result, Y, Z).


% The execute method first checks to see if the number of bulls is 5.
% If not it calls methods to create rules based on the number of bulls and cows.
execute(_, 5, _) :- write('Thanks for playing!'), clean.
execute(Guess, Bulls, Cows) :- handle(Guess, Bulls, Cows),
                               main.

% This method calls the method to generate guesses and get the number of bulls and cows from the user.
guesses(Result, Y, Z) :- retract(guessnum(X)),
                   write('Guess Number '), write(X), nl,
                   A is X + 1, assert(guessnum(A)),
                   gen_guess(5, [], Result), write('My guess is: '),
                   write(Result), nl,
                   write('Please enter the number of bulls followed by a period: '),
                   read(Y), nl,
                   write('Please enter the number of cows followed by a period: '),
                   read(Z), nl.

% This method generates two color guesses.
gen_guess(0, Guess, Guess) :- retract(available([_|T])),
			      assert(available(T)).
gen_guess(5, Guess, Result) :- not(numinferences(5)),
			       retract(available([H|T])),
			       assert(available(T)),
			       retract(beingconsidered(_)),
			       assert(beingconsidered(H)),
			       gen_guess(4, [H|Guess], Result).
gen_guess(Pos, Guess, Result) :- not(numinferences(5)),
				 available([H|_]),
				 Pos2 is Pos - 1,
				 retract(beingfixed(_)),
				 assert(beingfixed(H)),
				 gen_guess(Pos2, [H|Guess], Result).

% This method makes certain inferences about which colors are in the code.  There is some estimation involved here.
handle(_, 0, 0).
handle(Guess, Bulls, 0) :- retract(numinferences(R)),
			   New is R + Bulls,
			   assert(numinferences(New)),
			   beingconsidered(Y),
			   retract(others(A)),
			   handleaddrules(Bulls, Guess),
			   assert(rules(Y, [1, 2, 3, 4, 5])),
			   assert(others([Y|A])).
handle(_, 0, 1) :- retract(numinferences(R)),
		       New is R + 1,
		       assert(numinferences(New)),
		       beingfixed(X),
		       beingconsidered(Y),
		       retract(colors(Z)),
		       assert(colors([Y|Z])),
		       retract(others(A)),
		       assert(rules(X, [1, 2, 3, 4, 5])),
		       assert(rules(Y, [1, 2, 3, 4, 5])),
		       assert(others([X|A])).
handle(Guess, Bulls, 1) :- retract(numinferences(R)),
                           New is R + Bulls + 1,
                           assert(numinferences(New)),
			   beingfixed(X),
			   beingconsidered(Y),
			   retract(colors(Z)),
			   retract(others(A)),
			   assert(others([Y|A])),
			   assert(colors([Y|Z])),
			   handleaddrules(Bulls, Guess),
			   assert(rules(X, [1, 2, 3, 4, 5])),
			   assert(rules(Y, [1, 2, 3, 4, 5])).
handle(Guess, Bulls, 2) :- retract(numinferences(W)),
                           New is W + Bulls + 2,
                           assert(numinferences(New)),
			   beingfixed(X),
                           beingconsidered(Y),
			   retract(others(A)),
			   assert(others([Y|A])),
                           handleaddrules(Bulls, Guess),
			   assert(rules(Y, [1, 2, 3, 4])),
                           retract(colors(Z)),
                           append([X], Z, R),
			   append([Y], R, T),
			   assert(colors(T)), 
			   retract(numtied(B)),
		     	   C is B + 1,
		           assert(numtied(C)),
		           assert(tied(X, 5)).

% In this phase we want to see if our inferences about the colors in the code are correct.
% If the combination of bulls and cows comes out to five we can move into phase3 if not we have
% to make some substitutions.
phase2 :- numinferences(X),
	  X is 5,
	  retract(beingfixed(_)),
	  assert(beingfixed(0)),
	  colors(Y),
	  retract(guessnum(A)),
	  write('Guess number '), write(A), nl,
	  W is A + 1,
	  assert(guessnum(W)),
	  write('My guess is '), write(Y), nl,
	  write('Please enter the number of bulls followed by a period: '),
          read(R), nl,
          write('Please enter the number of cows followed by a period: '),
          read(Z), nl,
	  execute2(Y, R, Z). 

phase2 :- fillincolors.

% This method tries to find out if we have the correct colors in the code.  If we do it goes to the next phase.
% If not it keeps trying different replacements until it finds the correct colors or reaches 12 guesses.  
execute2(_, _, _) :- guessnum(13),
		     write('I could not guess your code').
execute2(_, 5, _) :- write('Thanks for playing').
execute2(_, Bulls, Cows) :- Z is Bulls + Cows,
			    Z is 5,
			    colors(X), cleanrules(X),
			    phase3.
execute2(_, Bulls, Cows) :- curpos(6), findandreplace(Bulls, Cows),
                            L is Bulls + Cows,
                            retract(numcorrect(_)),
                            assert(numcorrect(L)),
			    phase2.
execute2(_, Bulls, Cows) :- findandreplace2(Bulls, Cows),
                            L is Bulls + Cows,
                            retract(numcorrect(_)),
                            assert(numcorrect(L)),
			    phase2.

% This method makes switches of the colors in the current color list in case our initial estimations were wrong.
findandreplace2(Bulls, Cows) :- Y is Bulls + Cows,
				numcorrect(X),
				Y > X,
				retract(curpos(B)),
				retract(colors(Z)),
				switch(B, 1, Z, Result),
			 	R is B + 1,
				assert(curpos(R)),
				assert(colors(Result)).
findandreplace2(_, _) :- retract(curpos(X)),
			 Y is X - 1,
			 retract(colors(Z)),
			 switch(X, 1, Z, Result),
			 switch(Y, 1, Result, Result2),
			 R is X + 1,
			 assert(curpos(R)),
			 assert(colors(Result2)).

% This method is responsible for the actual switches
switch(0, _, _, _).
switch(Pos, Pos, [H|T], [Y|T]) :- opp(H, Y).
switch(Pos, Cur, [H|T], Result) :- Cur2 is Cur + 1,
				   switch(Pos, Cur2, T, Result2),
				   append([H], Result2, Result). 

% This method is responsible for trying colors in the other list.
findandreplace(Bulls, Cows) :- Y is Bulls + Cows,
			       numcorrect(A),
			       Y =< A,
			       retract(filledin(X)),
		  	       retract(others([H|T])),
			       assert(filledin(H)),
		  	       assert(others(T)),
			       retract(colors(Z)),
			       switch(5, 1, Z, Result),
			       assert(colors(Result)),
			       numadd(B),
			       remall(B, X),
			       retract(beingfixed(_)),
			       assert(beingfixed(H)),
			       handleaddrules(B, H).


% If there are not five colors in the colors list we get the first color from the others list,
% we get the first color from others and fill in the spaces.

fillincolors :- colors(X),
		length(X, 0, Length),
		Numadd is 5 - Length,
		retract(others([H|T])),
	        assert(others(T)),
		retract(filledin(_)),
		assert(filledin(H)),
		addrules(Numadd, H),
		retract(numinferences(_)),
		assert(numinferences(5)),
		retract(numadd(0)),
		assert(numadd(Numadd)),
		phase2.

% This method is used by fillincolors to add extra rules and colors to the colors list.
addrules(0, _).
addrules(Numadd, Color) :- assert(rules(Color, [1, 2, 3, 4, 5])),
			   retract(colors(X)),
			   assert(colors([Color|X])),
			   Numadd2 is Numadd - 1,
			   addrules(Numadd2, Color).

% Returns the length of a given list
length([], Length, Length).
length([_|T], Temp, Length) :- Temp2 is Temp + 1, length(T, Temp2, Length).

% This method adds colors to the colors list and calls addlists to create rules for colors
handleaddrules(0, _).
handleaddrules(Numrules, Guess) :- beingfixed(Z),
                                   assert(rules(Z, [1, 2, 3, 4, 5])),
                                   Numrules2 is Numrules - 1,
                                   handleaddrules(Numrules2, Guess),
				   retract(colors(X)),
				   assert(colors([Z|X])).

% This method is called after we know we have the correct amount of colors in the inference list.
% It tries to find the correct positions for the colors
phase3 :- getbeingfixed,
	  colors(X),
	  getbeingconsidered(X),
	  guesses3(Result, Y, Z),
	  execute3(Result, Y, Z).

% This method checks to see if all the colors are right. If not it makes inferences based on the guess.
execute3(_, 5, _) :- write('Thanks for playing!').
execute3(_, _, _) :- guessnum(13), write('I could not guess your code').
execute3(_, _, Cows) :- numcows(Cows),
			   phase3.
	  
% This method calls the gen_guess method to generate a new guess with guess number and gets the number of bulls and cows from the user.
guesses3(Result, Y, Z) :- retract(guessnum(X)),
                   write('Guess Number '), write(X), nl,
                   A is X + 1, assert(guessnum(A)),
                   gen_guess3(5, [], Result), write('My guess is: '),
                   write(Result), nl,
                   write('Please enter the number of bulls followed by a period: '),
                   read(Y), nl,
                   write('Please enter the number of cows followed by a period: '),
                   read(Z), nl.

%This method uses what we already know about rules and number of colors tied to generate new guesses
gen_guess3(0, Guess, Guess).
gen_guess3(Pos, Temp, Guess) :- tied(X, Pos),
                         Pos2 is Pos - 1,
                         gen_guess3(Pos2, [X|Temp], Guess).
gen_guess3(Pos, Temp, Guess) :- not(beingfixed(0)),
                         beingfixed(X),
                         rules(X, [H|_]),
                         H is Pos,
                         Pos2 is Pos-1,
                         gen_guess3(Pos2, [X|Temp], Guess).
gen_guess3(Pos, Temp, Guess) :- beingconsidered(X),
                         Pos2 is Pos - 1,
                         gen_guess3(Pos2, [X|Temp], Guess).

% This method makes the actual inferences
numcows(0) :- retract(beingfixed(X)),
              retract(rules(X, [H|_])),
              assert(tied(X, H)),
              retract(numtied(Z)),
              Y is Z + 1,
              assert(numtied(Y)),
              cleanup(H),
              cleanupall,
              assert(beingfixed(0)).
numcows(1) :- beingfixed(X),
              rules(X, [H|_]),
	      remall2(X, H).              
numcows(2) :- beingfixed(X),
              rules(X, [H|_]),
              beingconsidered(Y),
              assert(tied(Y, H)),
              retract(rules(Y,_)),
              retract(colors(Z)),
	      retract(numtied(R)),
	      L is R + 1,
	      assert(numtied(L)),
              select(Y, Z, Result),
              assert(colors(Result)),
              cleanup(H),
              cleanupall.

% This method removes the first instance of a given color from the inference list.
remfirst(_, [], []).
remfirst(Pos, [Pos|T], T).
remfirst(Pos, [_|T], Result) :- remfirst(Pos, T, Result).

%This method is used by cleanup to remove all instances of a given position from a given rule
rempos(_,[],[]).
rempos(Pos, [Pos|T], Result) :- rempos(Pos, T, Result).
rempos(Pos, [H|T], [H|TResult]) :- rempos(Pos, T, TResult).

% This method removes a given position from all rules of a given color.
remall2(Color, Pos) :- retract(rules(Color, X)),
                      rempos(Pos, X, Result),
                      remall2(Color, Pos),
                      asserta(rules(Color, Result)).
remall2(_, _).

% This method removes a certain color from the colors list.
remall(0, _).
remall(Numrem, Color) :- retract(colors(X)),
			 select(Color, X, Result),
			 assert(colors(Result)),
			 Numrem2 is Numrem-1,
			 remall(Numrem2, Color).

% This method gets a new value for beingfixed if necessary
getbeingfixed :- beingfixed(X),
		 not(X is 0).
getbeingfixed :- retract(colors([H|T])),
		 assert(colors(T)),
		 retract(beingfixed(0)),
		 assert(beingfixed(H)).

% This method gets the next color for being considered.  It takes the next color from the colors list.
getbeingconsidered([]).
getbeingconsidered([H|T]) :- beingfixed(Y),
		      	     H=Y,
		      	     getbeingconsidered(T).
getbeingconsidered([H|_]) :- retract(beingconsidered(_)),
		      	     assert(beingconsidered(H)).

%This method removes a given position from all rules in the database
cleanup(Pos) :- retract(rules(X,Y)),   
                rempos(Pos, Y, Result),
                cleanup(Pos),
                asserta(rules(X,Result)).
cleanup(_).

% ties colors to positions and removes those positions from all other rules
cleanupall :- rules(X,[H|T]),
              length([H|T], 0, Length),
              Length is 1,
              retract(rules(X, [H|T])),
              assert(tied(X,H)),
              retract(numtied(Z)),
              R is Z + 1, 
              assert(numtied(R)),
              retract(colors(Y)),
              rempos(X, Y, Result),
              assert(colors(Result)), 
              cleanup(H),
              cleanupall.
cleanupall.

cleanrules([]) :- retractrest.
cleanrules([H|T]) :- retract(rules(H, X)),
		    cleanrules(T),
		    assert(rules(H, X)).
cleanrules([H|T]) :- retract(colors(X)),
		     select(H, X, Result),
		     assert(colors(Result)),
		     cleanrules(T).

% retracts all rules
retractrest :- retract(rules(_,_)),
	       retractrest.
retractrest.

% retracts all tied colors
retractied :- retract(tied(_, _)),
	      retractied.
retractied.

% This method cleans up all the variables so the user doesn't have to restart prolog every time they want to play
clean :- retractrest,
	 retractied,
         retract(guessnum(_)),
         assert(guessnum(1)),
         retract(available(_)),
         assert(available([1, 2, 3, 4, 5, 6, 7, 8])),
         retract(colors(_)),
         assert(colors([])),
         retract(numinferences(_)),
         assert(numinferences(0)),
         retract(numtied(_)),
         assert(numtied(0)),
         retract(beingfixed(_)),
         assert(beingfixed(0)),
         retract(beingconsidered(_)),
         assert(beingconsidered(0)),
	 retract(others(_)),
	 assert(others([])),
	 retract(filledin(_)),
	 assert(filledin(0)),
	 retract(curpos(_)),
	 assert(curpos(1)),
	 retract(numcorrect(_)),
	 assert(numcorrect(0)),
	 retract(numadd(_)),
	 assert(numadd(0)).

% dynamic variables
available([1, 2, 3, 4, 5, 6, 7, 8]).
numinferences(0).
numtied(0).
beingfixed(0).
beingconsidered(0).
guessnum(1).
colors([]).
others([]).
filledin(0).
numadd(0).

opp(1, 2).
opp(2, 1).
opp(3, 4).
opp(4, 3).
opp(5, 6).
opp(6, 5).
opp(7, 8).
opp(8, 7).

curpos(1).
numcorrect(0).
